-- functions extracted from neil.e Pete Eberlein
include machine.e
include image.e
------------------------------------------------------------------------
-- FONTS based on Jiri Babor's fonts
------------------------------------------------------------------------

global constant
    BOLD = 1,
    ITALIC = 2, 
    UNDERLINE = 4,
    DIMMED = 8

sequence char, char_widths
integer char_height, baseline_height, text_x, text_y
char = {}
text_x = 0
text_y = 0

sequence all_fonts
all_fonts = {}

global procedure rom_font(integer width, integer height)
--> select a rom font
-->  width may be 8 or 9 
-->  height may be 8, 14 or 16
--> Note: width 9 doesn't work with height 8
    sequence regs, charaddr
    integer font_buf, bits, on, mask, x

    regs = repeat(0, 10)
    regs[REG_AX] = #1130
    if width = 8 and height = 8 then
        regs[REG_BX] = #0300  
    elsif (width = 8 or width = 9) and height = 14 then
        regs[REG_BX] = #0200
    elsif (width = 8 or width = 9) and height = 16 then
        regs[REG_BX] = #0600
    else
        return
    end if

    regs = dos_interrupt(#10, regs)
    font_buf = regs[REG_ES] * #10 + regs[REG_BP]
    charaddr = repeat(0, 256)
    for c = 1 to 256 do
        charaddr[c] = c * height
    end for
    charaddr = charaddr + font_buf

    char_widths = repeat(width, 256)
    char_height = height
    baseline_height = height

    if width = 9 then
        if height = 14 then
            regs[REG_BX] = #0500
        elsif height = 16 then
            regs[REG_BX] = #0700
        end if
        regs = dos_interrupt(#10, regs)
        font_buf = regs[REG_ES] * #10 + regs[REG_BP]
        while peek(font_buf) do
            charaddr[peek(font_buf)] = font_buf + 1
            font_buf = font_buf + 1 + height
        end while
    end if
            
    char = repeat({}, 256)
    for newAscii = 1 to 256 do
        for y = 1 to height do
            bits = peek(charaddr[newAscii] + y-1)
            on = 0  x = 8  mask = 1
            while mask < bits do
                if and_bits(bits, mask) then
                    if on then
                        char[newAscii][on] = char[newAscii][on] + 1
                        char[newAscii][on-2] = x
                    else
                        char[newAscii] = char[newAscii] & {x,y,0}
                        on = length(char[newAscii])
                    end if
                elsif on then
                    on = 0
                end if
                x = x - 1
                mask = mask * 2
            end while
        end for
    end for
end procedure

global procedure add_pitch(integer pitch)
--> add pitch to horizontal font character spacing
    char_widths = char_widths + pitch
end procedure


global function build_text_image(sequence chars, integer foreground, object background, integer attributes)
--> returns a 2-d image sequence of chars
--> foreground and background are colors, same format as pixel, get_pixel, etc.
--> attributes is an or'ed value of BOLD, ITALIC, UNDERLINE, or DIMMED
    integer x, y, l, dx, fbold
    sequence ch, image, ioffset
    
    if not length(char) then
        rom_font(8,16)        
    end if
    
    ioffset = repeat(0, char_height)
    if and_bits(attributes, ITALIC) then
        for i = 1 to char_height do
            ioffset[i] = floor((char_height - i) * 0.25 )
        end for
    end if

    fbold = and_bits(attributes, BOLD) != 0
    l = 0
    for i = 1 to length(chars) do
        l = l + char_widths[chars[i]] + fbold
    end for
    image = repeat(repeat(background, l+ioffset[1]), char_height)

    if and_bits(attributes, UNDERLINE) then
        x = 1 + ioffset[baseline_height]
        image[baseline_height][x..x+l-1] = foreground
    end if
    
    if and_bits(attributes, DIMMED) then
        dx = 0
        for c = 1 to length(chars) do
            ch = char[chars[c]]
            for i = 1 to length(ch)-2 by 3 do
                y = ch[i+1]
                x = ch[i] + dx + ioffset[y]
                for xx = x+and_bits(x+y,1) to x+ch[i+2]+fbold by 2 do
                    image[y][xx] = foreground
                end for
            end for
            dx = dx + char_widths[chars[c]] + fbold
        end for
    else
        dx = 0
        for c = 1 to length(chars) do
            ch = char[chars[c]]
            for i = 1 to length(ch)-2 by 3 do
                y = ch[i+1]
                x = ch[i] + dx + ioffset[y]
                image[y][x..x+ch[i+2]+fbold] = foreground
            end for
            dx = dx + char_widths[chars[c]] + fbold
        end for
    end if
    return image
end function




