--Base128.e written by Paul Plummer
--Base 128 encoding of binary data
--all numbers use least significant byte first notation

--converts binary data into printable charaters that can be read by euphoria
include machine.e
integer pow24,pow21,pow14,pow7 
sequence Base128_chars
Base128_chars=    --skip escape character '\'
"0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnop"&
"qrstuvwxyz{|}~"
pow24=power(2,24)
pow21=power(2,21)
pow14=power(2,14)
pow7= power(2,7)


global function bytes_to_base128(sequence input)
--read 7 byte chunks as a 56 bit number - bit56_num
--split into two 28 bit chunks for processing  - lchunk mchunk
--     7 bytes in --> 8 printable characters out
sequence bit56_num,junk,base128,output,ochunk
integer lchunk,mchunk   -- 28 bit chunks
integer lbyte,mbyte  -- 4 bit partial chunks
integer last_chunk,num_chunks
last_chunk=remainder(length(input),7)
num_chunks=floor(length(input)/7)
base128=repeat(0,8)
output={Base128_chars[last_chunk+1]}    --first character = length last_chunk
for c=1 to num_chunks do
bit56_num=input[(c-1)*7+1..(c-1)*7+7]     --read 7 bytes 56 bits
junk=bit56_num[1..3]&0    --read 24 bits of lchunk
lchunk=bytes_to_int(junk)
mbyte=floor(bit56_num[4]/16)     -- 4 bit partial
lbyte=remainder(bit56_num[4],16) -- 4 bit partial 
lchunk+=pow24*lbyte           --create 28 bit lchunk
junk=bit56_num[5..7]&0    --read 24 bits of mchunk    
mchunk=bytes_to_int(junk) 
mchunk*=16
mchunk+=mbyte                --create 28 bit mchunk
--convert to base 128
base128[8]=floor(mchunk/pow21)
mchunk=remainder(mchunk,pow21)
base128[7]=floor(mchunk/pow14)
mchunk=remainder(mchunk,pow14)
base128[6]=floor(mchunk/pow7)
mchunk=remainder(mchunk,pow7)
base128[5]=mchunk
base128[4]=floor(lchunk/pow21)
lchunk=remainder(lchunk,pow21)
base128[3]=floor(lchunk/pow14)
lchunk=remainder(lchunk,pow14)
base128[2]=floor(lchunk/pow7)
lchunk=remainder(lchunk,pow7)
base128[1]=lchunk
ochunk={}
for cc=1 to 8 do    --convert to printable characters
ochunk&=Base128_chars[base128[cc]+1]
end for
output&=ochunk     --add to output
end for
if last_chunk then  --process last partial chunk - less than 7 bytes
bit56_num=input[(num_chunks)*7+1..length(input)]
--add padding and process as before
for c=1 to 7-length(bit56_num) do
bit56_num&=0  --padding
end for
junk=bit56_num[1..3]&0
lchunk=bytes_to_int(junk)
mbyte=floor(bit56_num[4]/16)
lbyte=remainder(bit56_num[4],16)
lchunk+=pow24*lbyte
junk=bit56_num[5..7]&0
mchunk=bytes_to_int(junk) 
mchunk*=16
mchunk+=mbyte
base128[8]=floor(mchunk/pow21)
mchunk=remainder(mchunk,pow21)
base128[7]=floor(mchunk/pow14)
mchunk=remainder(mchunk,pow14)
base128[6]=floor(mchunk/pow7)
mchunk=remainder(mchunk,pow7)
base128[5]=mchunk
base128[4]=floor(lchunk/pow21)
lchunk=remainder(lchunk,pow21)
base128[3]=floor(lchunk/pow14)
lchunk=remainder(lchunk,pow14)
base128[2]=floor(lchunk/pow7)
lchunk=remainder(lchunk,pow7)
base128[1]=lchunk
ochunk={}
for cc=1 to 8 do
ochunk&=Base128_chars[base128[cc]+1]
end for
output&=ochunk
end if
return output
end function

global function base128_to_bytes(sequence input)
--read 8 character chunks - 56 bit number encoded as base 128
--convert back to original 7 byte binary 
--   8 characters in --> 7 bytes out
sequence junk,base128,output,ochunk
integer lchunk,mchunk   --28 bit chunks
integer lbyte,mbyte,last_chunk,num_chunks
last_chunk=input[1]-48
num_chunks=floor(length(input)/8)
output={}
for cc=1 to num_chunks do    --read 8 characters
base128=input[(cc-1)*8+2..(cc-1)*8+9] 
base128-=48                 --convert to powers

for c=1 to 8 do            --handle skipped '\' character
if base128[c]>44 then
base128[c]-=1    
end if
end for
--create 28 bit chunks
lchunk=base128[1]+base128[2]*128+base128[3]*16384+base128[4]*2097152
mchunk=base128[5]+base128[6]*128+base128[7]*16384+base128[8]*2097152
ochunk=repeat(0,7)
lbyte=floor(lchunk/pow24)       -- 4 bit partial
lchunk=remainder(lchunk,pow24)  -- 24 bit partial
mbyte=remainder(mchunk,16)      -- 4 bit partial 
mchunk=floor(mchunk/16)         -- 24 bit partial
junk=int_to_bytes(lchunk)
ochunk[1]=junk[1]
ochunk[2]=junk[2]
ochunk[3]=junk[3]
ochunk[4]=lbyte+mbyte*16       --combine 4 bit partials
junk=int_to_bytes(mchunk)
ochunk[5]=junk[1]
ochunk[6]=junk[2]
ochunk[7]=junk[3]
output&=ochunk                --add to output
end for
if last_chunk then      --remove any padding bytes
output=output[1..length(output)-7+last_chunk]
end if
return output
end function


