--reads and writes palatted non_interlaced non_prefilterd png images

include get.e
include machine.e
include misc.e
include crc32.e
include inflate.e
include zdata.e
integer fn_input,fn_output
sequence png_sig
png_sig={137,80,78,71,13,10,26,10}
function read_chunk()
sequence chunk_length,chunk_name,chunk_dat,chunk_crc
integer ch_length,crc_ok
chunk_length=get_bytes(fn_input,4)
if length(chunk_length)<4 then
return {32}  -- truncated file
end if
ch_length=bytes_to_int(reverse(chunk_length))
chunk_name=get_bytes(fn_input,4)
if length(chunk_name)<4 then
return {32} --bad chunk name
end if
chunk_dat=get_bytes(fn_input,ch_length)
if length(chunk_dat)<ch_length then
return {32} -- truncated file
end if
chunk_crc=reverse(get_bytes(fn_input,4))
if length(chunk_crc)<4 then
return {32}  --truncated file
end if
crc_ok=compare(crc_32(chunk_name&chunk_dat),chunk_crc)
if not(crc_ok=0) then
return {33}  --bad chunk crc
end if
return {chunk_name,chunk_dat}
end function

function save_chunk(sequence chunk_name,sequence chunk_dat)
sequence chunk_crc,chunk_out,chunk_length
chunk_crc=reverse(crc_32(chunk_name&chunk_dat))
chunk_length=reverse(int_to_bytes(length(chunk_dat)))
chunk_out=chunk_length&chunk_name&chunk_dat&chunk_crc
return chunk_out
end function

procedure save_plte(sequence pal)
sequence pal_out,dat_out
pal_out={}
for c=1 to length(pal) do
pal_out&=pal[c]
end for
dat_out=save_chunk("PLTE",pal_out)
puts(fn_output,dat_out)
end procedure

procedure save_ihdr(sequence image)
sequence ihdr_dat,dat_out
ihdr_dat=reverse(int_to_bytes(length(image[1])))
ihdr_dat&=reverse(int_to_bytes(length(image)))
ihdr_dat&={8,3,0,0,0}
dat_out=save_chunk("IHDR",ihdr_dat)
puts(fn_output,dat_out)
end procedure

function save_idat(sequence image)
sequence idat_comp,dat_out
integer fn_temp
fn_temp=open("idat.tmp","wb")
if fn_temp=-1 then
EZ_ERROR=36
return EZ_ERROR
end if
for c=1 to length(image) do
puts(fn_temp,0)
puts(fn_temp,image[c])
end for
close(fn_temp)
idat_comp=deflate_seq("idat.tmp",0)
if length(idat_comp)=0 then
return EZ_ERROR
end if
idat_comp=idat_comp[10..length(idat_comp)]
--idat_comp={120,156}&idat_comp[1..length(idat_comp)-4]
idat_comp={120,156}&idat_comp

dat_out=save_chunk("IDAT",idat_comp)
system("del idat.tmp",2) 
puts(fn_output,dat_out)
return 0
end function

global function save_png(sequence pal_image,sequence fname)
sequence dat_out
integer test
fn_output=open(fname,"wb")
if fn_output=-1 then
EZ_ERROR=31
return -1
end if
puts(fn_output,png_sig)
save_ihdr(pal_image[2])
save_plte(pal_image[1])
test=save_idat(pal_image[2])
if test then
close(fn_output)
system("del "&fname,2) 
return test
end if
dat_out=save_chunk("IEND",{})
puts(fn_output,dat_out)
close(fn_output)
EZ_ERROR=0
return 0
end function
--            IHDR format
--            Width:              4 bytes
--            Height:             4 bytes
--            Bit depth:          1 byte
--            Color type:         1 byte
--            Compression method: 1 byte
--            Filter method:      1 byte
--            Interlace method:   1 byte

global integer png_width,png_height,png_bits,png_color,png_comp
global integer png_filter,png_interlace
global sequence png_pal,png_image
global integer png_pal_count
procedure ihdr(sequence ihdr_dat)
png_width=bytes_to_int(reverse(ihdr_dat[1..4]))
png_height=bytes_to_int(reverse(ihdr_dat[5..8]))
png_bits=ihdr_dat[9]
png_color=ihdr_dat[10]
png_comp=ihdr_dat[11]
png_filter=ihdr_dat[12]
png_interlace=ihdr_dat[13]
end procedure

procedure open_file(sequence fname)
sequence fsig
fn_input=open(fname,"rb")
if fn_input>0 then
 fsig=get_bytes(fn_input,8)
  if compare(png_sig,fsig)=0 then
EZ_ERROR=0
  else
   close(fn_input)
EZ_ERROR=34  --not a png file
  end if
else
EZ_ERROR=31   --file not found
end if
end procedure

global procedure info_png(sequence fname)
sequence test
open_file(fname)
if EZ_ERROR then
return
end if

while 1 do
test=read_chunk()
if length(test)=1 then
EZ_ERROR=test[1]
exit
end if
if compare(test[1],"IHDR")=0 then
ihdr(test[2])
exit
end if
end while
close(fn_input)
end procedure

global function read_png(sequence fname)
integer offs,offe
sequence temp_line
sequence test,junk,idat
idat={} 
open_file(fname)
if EZ_ERROR then
return EZ_ERROR
end if
while 1 do
test=read_chunk()
if length(test)=1 then 
EZ_ERROR=test[1]
return EZ_ERROR
end if
if compare(test[1],"IEND")=0 then exit end if

if compare(test[1],"IHDR")=0 then
ihdr(test[2])
if png_bits<9 and png_color=3 and png_comp=0 and png_filter=0 and png_interlace=0 then
else
close(fn_input)
EZ_ERROR=35
return EZ_ERROR  --unsupported format
end if
end if
if compare(test[1],"IDAT")=0 then
idat&=test[2]
end if
if compare(test[1],"PLTE")=0 then
png_pal_count=length(test[2])/3
png_pal=repeat({0,0,0},png_pal_count)
offs=1
offe=3
for c=1 to png_pal_count do
png_pal[c]=test[2][offs..offe]
offs+=3
offe+=3
end for
end if
end while
close(fn_input)
junk=inflate(idat[3..length(idat)],(png_width+1)*png_height)
if junk[1] then
EZ_ERROR=junk[1]
return EZ_ERROR
end if     
png_image={}
offs=1
offe=png_width+1
for c=1 to png_height do
temp_line=junk[2][offs..offe]
png_image=append(png_image,temp_line[2..length(temp_line)])
offs+=png_width+1
offe+=png_width+1
end for
return {png_pal,png_image}
end function

global function PNG_ERROR_STRING()
if EZ_ERROR=0 then
return "No errors"
else
return EZ_ERROR_STRING[EZ_ERROR]    
end if
end function

global function PNG_ERROR()
return EZ_ERROR    
end function

global procedure CLEAR_PNG_ERROR()
EZ_ERROR=0    
end procedure

