-----------------------------
-- BGET.E             v1.0 --
-- (c) 1999 Gabriel Boehme --
-- requires  Euphoria 2.1+ --
-----------------------------
--modifications by Paul Plummer
--renamed bprint to zd_bprint to avoid conflicts if bget.e is included
--sbget() - modified bget function that reads data from a sequence instead
--          of a file
--zdm_bprint() - modified bprint function that outputs data to a sequence
include get.e
include machine.e
sequence seq  --takes the place of a input file in bget.e
integer count --keeps track of position in data
sequence sbget_out --takes place of output file in bget.e

global procedure init_sbget(sequence inseq)  --reset input data
seq=inseq
count=1
end procedure

global procedure close_sbget()   --free input data memory
seq={}
end procedure

global procedure clear_sbget_out()   --free input data memory
sbget_out={}
end procedure

function get_char()   --getc(fn) 
integer char 
if count>length(seq) then
return -1    
end if
char=seq[count]
count+=1
return char
end function

function get_chars(integer num_char)   --get_bytes(fn,num_char) 
sequence sq
if (count+num_char-1)>length(seq)  then
return -1
end if
sq=seq[count..count+num_char-1]
count+=num_char
return sq
end function

global type string8(sequence s) -- string8
object x
   for i = 1 to length(s) do
      x = s[i]
      if integer(x) then
         if x < 0 then
            return 0
         elsif x > #FF then
            return 0
         end if
      else
         return 0
      end if
   end for
   return 1
end type

integer fn, stype, atype, byte
atom sign, a

constant ZERO = 0, BYTE = 1, INT16 = 2, INT32 = 4, FLOAT64 = 8
                -- -BYTE #F, -INT16 #E, -INT32 #C
constant NONE = 0, SEQUENCE = 1, STRING = 2

procedure Print(object x)
integer len
sequence s
   if atom(x) then
      sign = compare(x, 0) --sign
      a = x * sign --absolute value
      if integer(a) then
         if a = 0 then
            atype = ZERO
            s = {}
         elsif a <= #FF then
            atype = BYTE * sign
            s = {a}
         elsif a <= #FFFF then
            atype = INT16 * sign
            s = {remainder(a,#100), floor(a/#100)}
         else
            atype = INT32 * sign
            s = int_to_bytes(a)
         end if
      elsif a = floor(a) and a <= #FFFFFFFF then
         atype = INT32 * sign
         s = int_to_bytes(a)
      else
         atype = FLOAT64
         s = atom_to_float64(x)
      end if
      if atype < 0 then
         atype += #10
      end if
      if stype then
         atype += stype * #10
         stype = NONE
      end if
      puts(fn, prepend(s, atype))
   else
      s = x
      len = length(s)
      if string8(s) then
         stype = STRING
         Print(len)
         puts(fn, s)
      else
         stype = SEQUENCE
         Print(len)
         for i = 1 to len do
            Print(s[i])
         end for
      end if
   end if
end procedure

global procedure zd_bprint(integer file, object x)    --same as bprint()
   fn = file
   stype = NONE
   Print(x)
end procedure


global function sbget()  --same as bget() but reads from a sequence seq
integer len              --instead of a file
sequence s, g
   byte = get_char()     --getc(fn)
   if byte = -1 then
      return {GET_EOF}
   end if
      
   atype = remainder(byte, #10)
   stype = floor(byte/#10)
   
   if atype > FLOAT64 then
      atype = #10 - atype
      sign = -1
   else
      sign = 1
   end if
   
   if atype = ZERO then
      a = 0
   elsif atype = BYTE then
      a = get_char()       --getc(fn
      if a = -1 then
         return {GET_EOF, 0}
      end if
      a *= sign
   elsif atype = INT16 then
      s = get_chars(2)   --get_bytes(fn,2)
      if atom(s) then
         return {GET_EOF, s}
      end if
      a = (s[1] + s[2] * #100) * sign
   elsif atype = INT32 then
      s = get_chars(4)   --get_bytes(fn,4)
       if atom(s) then
         return {GET_EOF, s}
        end if
      a = bytes_to_int(s) * sign
   elsif atype = FLOAT64 then
      s = get_chars(8)   --get_bytes(fn,8)
      if atom(s) then
         return {GET_EOF, s}
      end if
      a = float64_to_atom(s)
   else
      return {GET_FAIL, byte}
   end if
   
   if stype then
      len = a
      if stype = STRING then
         s = get_chars(len)   --get_bytes(fn,len)
          if atom(s) then
            return {GET_EOF, s}
         end if
      elsif stype = SEQUENCE then
         s = repeat(0, len)
         for i = 1 to len do
            g = sbget()
            if g[1] != GET_SUCCESS then
               return g
            end if
            s[i] = g[2]
         end for
      else
         return {GET_FAIL, byte}
      end if
      
      return {GET_SUCCESS, s}
   else
      return {GET_SUCCESS, a}
   end if
end function

procedure zdm_Print(object x)
integer len
sequence s
   if atom(x) then
      sign = compare(x, 0) --sign
      a = x * sign --absolute value
      if integer(a) then
         if a = 0 then
            atype = ZERO
            s = {}
         elsif a <= #FF then
            atype = BYTE * sign
            s = {a}
         elsif a <= #FFFF then
            atype = INT16 * sign
            s = {remainder(a,#100), floor(a/#100)}
         else
            atype = INT32 * sign
            s = int_to_bytes(a)
         end if
      elsif a = floor(a) and a <= #FFFFFFFF then
         atype = INT32 * sign
         s = int_to_bytes(a)
      else
         atype = FLOAT64
         s = atom_to_float64(x)
      end if
      if atype < 0 then
         atype += #10
      end if
      if stype then
         atype += stype * #10
         stype = NONE
      end if
--      puts(fn, prepend(s, atype))
sbget_out&=prepend(s,atype)
   else
      s = x
      len = length(s)
      if string8(s) then
         stype = STRING
         zdm_Print(len)
sbget_out&=s --         puts(fn, s)
      else
         stype = SEQUENCE
         zdm_Print(len)
         for i = 1 to len do
            zdm_Print(s[i])
         end for
      end if
   end if
end procedure

global function zdm_bprint(object x)    --same as bprint except prints to
                                        --a sequence
   stype = NONE
sbget_out={}
   zdm_Print(x)
return sbget_out
end function



